# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/01_repr_rgb.ipynb.

# %% auto 0
__all__ = ['rgb']

# %% ../nbs/01_repr_rgb.ipynb 3
from typing import Union, Any, Optional as O
from functools import cached_property

import numpy as np
from matplotlib import pyplot as plt, axes, figure, rcParams
from IPython.core.pylabtools import print_figure


from .utils.tile2d import hypertile
from .utils import get_config

# %% ../nbs/01_repr_rgb.ipynb 4
def fig_rgb(x           :np.ndarray,        # Array to display. [[...], C,H,W] or [[...], H,W,C]
            denorm      :Any    =None,      # Reverse per-channel normalizatoin
            cl          :Any    =True,      # Channel-last
            gutter_px   :int    =3,         # If more than one tensor -> tile with this gutter width
            frame_px    :int    =1,         # If more than one tensor -> tile with this frame width
            scale       :int    =1,         # Stretch the image. Only itegers please.
            view_width  :int    =966,       # target width of the image 
            clip        :bool   =True,      # Selently clip RGB values to [0, 1]
            ax          :O[axes.Axes]=None  # Matplotlib axes
        ) -> figure.Figure:
    
    assert x.ndim >= 3, f"Expecting 3 or more dimension input, got shape=({x.shape})"
    assert x.size > 0, f"Expecting non-empty input, got shape=({x.shape})"
    # swap channels if it's not channe-last already
    if not cl:
        x = np.swapaxes(np.swapaxes(x, -3, -1), -3, -2)

    scale = int(scale)
    x = x.repeat(scale, axis=-2).repeat(scale, axis=-3)

    n_ch = x.shape[-1]
    assert n_ch in (3, 4), f"Expecting 3 (RGB) or 4 (RGBA) color channels, got {n_ch}, shape=({x.shape})"
    if denorm:
        means = np.array(denorm[0])
        stds = np.array(denorm[1])
        x = x * stds + means

    if x.ndim > 3:
        x = hypertile(  t=x,
                        gutter_px=gutter_px,
                        frame_px=frame_px,
                        view_width=view_width)

    if clip: np.clip(x, 0, 1, out=x)

    cfg = get_config()
    close = cfg.fig_close and not cfg.fig_show # Don't close if requested to show
    show = cfg.fig_show and ax is None # Don't show the figure if axes was provided

    if ax is None:
        fig = plt.figure(frameon=False, figsize=(x.shape[1] * 0.01, x.shape[0]*0.01) )
        if close: plt.close(fig)
        fig.set_dpi(100)

        ax = fig.add_axes([0,0,1,1])
        ax.set_axis_off()
        ax.set_xlim(0, x.shape[1]+1)

    ax.imshow(x, interpolation="none")
    if show: plt.show()

    return ax.figure

# %% ../nbs/01_repr_rgb.ipynb 5
class RGBProxy():
    """Flexible `PIL.Image.Image` wrapper"""
    
    def __init__(self, x:np.ndarray):
        assert x.ndim >= 3, f"Expecting at least 3 dimensions, got shape{x.shape}={x.size}"
        self.x=x
        self.params = dict( denorm      = None,
                            cl          = True,
                            gutter_px   = 3,     
                            frame_px    = 1,
                            scale       = 1,
                            view_width  = 966,
                            clip        = True,
                            ax          = None)

    def __call__(self,
                denorm      :Any    =None,
                cl          :Any    =None,
                gutter_px   :O[int] =None,
                frame_px    :O[int] =None,
                scale       :O[int] =None,
                view_width  :O[int] =None,
                clip        :Any    =None,
                ax          :O[axes.Axes]=None):
        
        self.params.update( { k:v for
                            k,v in locals().items()
                            if k != "self" and v is not None } )
        _ = self.fig # Trigger figure generation
        return self

    @cached_property
    def fig(self) -> figure.Figure:
        return fig_rgb(self.x, **self.params)

    def _repr_png_(self):
        return print_figure(self.fig, fmt="png", pad_inches=0,
            metadata={"Software": "Matplotlib, https://matplotlib.org/"})

# %% ../nbs/01_repr_rgb.ipynb 6
def rgb(x           :np.ndarray,        # Array to display. [[...], C,H,W] or [[...], H,W,C]
        denorm      :Any    =None,      # Reverse per-channel normalizatoin
        cl          :Any    =True,      # Channel-last
        gutter_px   :int    =3,         # If more than one tensor -> tile with this gutter width
        frame_px    :int    =1,         # If more than one tensor -> tile with this frame width
        scale       :int    =1,         # Stretch the image. Only itegers please.
        view_width  :int    =966,       # target width of the image 
        clip        :bool   =True,      # Selently clip RGB values to [0, 1]
        ax          :O[axes.Axes]=None  # Matplotlib axes
        ) -> RGBProxy:

    args = locals()
    del args["x"]
    
    return RGBProxy(x)(**args)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/03d_utils.config.ipynb.

# %% auto 0
__all__ = ['set_config', 'get_config', 'config']

# %% ../../nbs/03d_utils.config.ipynb 4
from copy import copy
from types import SimpleNamespace
from typing import Callable, Union, Optional, TypeVar
from contextlib import contextmanager

import numpy as np

# %% ../../nbs/03d_utils.config.ipynb 5
class Config(SimpleNamespace):
    "Config"
    def __init__(self,
            precision     = 3,    # Digits after `.`
            threshold_max = 3,    # .abs() larger than 1e3 -> Sci mode
            threshold_min = -4,   # .abs() smaller that 1e-4 -> Sci mode
            sci_mode      = None, # Sci mode (2.3e4). None=auto
            indent        = 2,    # Indent for .deeper()
            color         = True, # ANSI colors in text
            repr          = None, # Use func e.g. `lovely` for `repr(np.ndarray)`
            str           = None, # Use func e.g. `lovely` for `str(np.ndarray)`
            plt_seed      = 42,   # Sampling seed for `plot`
            fig_close     = True, # Close matplotlib Figure
            fig_show      = False,# Call `plt.show()` for `.plt`, `.chans` and `.rgb`
    ):
        super().__init__(**{k:v for k,v in locals().items() if k not in ["self", "__class__"]})

_defaults = Config()
_config = copy(_defaults)

# %% ../../nbs/03d_utils.config.ipynb 8
# Allows passing None as an argument to reset the 
class _Default():
    def __repr__(self):
        return "Ignore"
D = _Default()
Default = TypeVar("Default")

# %% ../../nbs/03d_utils.config.ipynb 9
def set_config( precision       :Optional[Union[Default,int]]     =D,
                threshold_min   :Optional[Union[Default,int]]     =D,
                threshold_max   :Optional[Union[Default,int]]     =D,
                sci_mode        :Optional[Union[Default,bool]]    =D,
                indent          :Optional[Union[Default,bool]]    =D,
                color           :Optional[Union[Default,bool]]    =D,
                repr            :Optional[Union[Default,Callable]]=D,
                str             :Optional[Union[Default,Callable]]=D,
                plt_seed        :Optional[Union[Default,int]]     =D,
                fig_close       :Optional[Union[Default,bool]]    =D,
                fig_show        :Optional[Union[Default,bool]]    =D
                ) -> None:

    "Set config variables"
    args = locals().copy()
    for k,v in args.items():
        if v != D:
            
            # set_config(repr=func)             -> Set both `repr` and `str`.
            # set_config(repr=func, str=None)   -> Set `repr`, unset `str``
            # set_config(str=func)              -> Set `str`` only, don't tuch `repr``
            # set_config(repr=None)             -> Unset `repr`and `str``
            # set_config(str=None)              -> Unset `str` only

            if k == "repr":
                np.set_string_function(v, True)
                if args["str"] == D:
                    np.set_string_function(v, False)
            if k == "str":
                np.set_string_function(v, False)

            if v is None:
                setattr(_config, k, getattr(_defaults, k))
            else:
                setattr(_config, k, v)

# %% ../../nbs/03d_utils.config.ipynb 10
def get_config():
    "Get a copy of config variables"
    return copy(_config)

# %% ../../nbs/03d_utils.config.ipynb 11
@contextmanager
def config( precision       :Optional[Union[Default,int]]     =D,
            threshold_min   :Optional[Union[Default,int]]     =D,
            threshold_max   :Optional[Union[Default,int]]     =D,
            sci_mode        :Optional[Union[Default,bool]]    =D,
            indent          :Optional[Union[Default,bool]]    =D,
            color           :Optional[Union[Default,bool]]    =D,
            repr            :Optional[Union[Default,Callable]]=D,
            str             :Optional[Union[Default,Callable]]=D,
            plt_seed        :Optional[Union[Default,int]]     =D,
            fig_close       :Optional[Union[Default,bool]]    =D,
            fig_show        :Optional[Union[Default,bool]]    =D
            ):
    "Context manager for temporarily setting config options"
    new_opts = { k:v for k, v in locals().items() if v != D}
    old_opts = copy(get_config().__dict__)


    try:
        set_config(**new_opts)
        yield
    finally:
        set_config(**old_opts)



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'master',
                'doc_baseurl': '/lovely-numpy',
                'doc_host': 'https://xl0.github.io',
                'git_url': 'https://github.com/xl0/lovely-numpy',
                'lib_path': 'lovely_numpy'},
  'syms': { 'lovely_numpy.lo': { 'lovely_numpy.lo.Lo': ('lo.html#lo', 'lovely_numpy/lo.py'),
                                 'lovely_numpy.lo.Lo.__call__': ('lo.html#lo.__call__', 'lovely_numpy/lo.py'),
                                 'lovely_numpy.lo.Lo.__init__': ('lo.html#lo.__init__', 'lovely_numpy/lo.py'),
                                 'lovely_numpy.lo.Lo.__repr__': ('lo.html#lo.__repr__', 'lovely_numpy/lo.py'),
                                 'lovely_numpy.lo.Lo.chans': ('lo.html#lo.chans', 'lovely_numpy/lo.py'),
                                 'lovely_numpy.lo.Lo.deeper': ('lo.html#lo.deeper', 'lovely_numpy/lo.py'),
                                 'lovely_numpy.lo.Lo.p': ('lo.html#lo.p', 'lovely_numpy/lo.py'),
                                 'lovely_numpy.lo.Lo.plt': ('lo.html#lo.plt', 'lovely_numpy/lo.py'),
                                 'lovely_numpy.lo.Lo.rgb': ('lo.html#lo.rgb', 'lovely_numpy/lo.py'),
                                 'lovely_numpy.lo.Lo.v': ('lo.html#lo.v', 'lovely_numpy/lo.py'),
                                 'lovely_numpy.lo.lo': ('lo.html#lo', 'lovely_numpy/lo.py')},
            'lovely_numpy.repr_chans': { 'lovely_numpy.repr_chans.ChanProxy': ('repr_chans.html#chanproxy', 'lovely_numpy/repr_chans.py'),
                                         'lovely_numpy.repr_chans.ChanProxy.__call__': ( 'repr_chans.html#chanproxy.__call__',
                                                                                         'lovely_numpy/repr_chans.py'),
                                         'lovely_numpy.repr_chans.ChanProxy.__init__': ( 'repr_chans.html#chanproxy.__init__',
                                                                                         'lovely_numpy/repr_chans.py'),
                                         'lovely_numpy.repr_chans.ChanProxy._repr_png_': ( 'repr_chans.html#chanproxy._repr_png_',
                                                                                           'lovely_numpy/repr_chans.py'),
                                         'lovely_numpy.repr_chans.ChanProxy.fig': ( 'repr_chans.html#chanproxy.fig',
                                                                                    'lovely_numpy/repr_chans.py'),
                                         'lovely_numpy.repr_chans.chans': ('repr_chans.html#chans', 'lovely_numpy/repr_chans.py'),
                                         'lovely_numpy.repr_chans.fig_chans': ('repr_chans.html#fig_chans', 'lovely_numpy/repr_chans.py')},
            'lovely_numpy.repr_plt': { 'lovely_numpy.repr_plt.PlotProxy': ('repr_plt.html#plotproxy', 'lovely_numpy/repr_plt.py'),
                                       'lovely_numpy.repr_plt.PlotProxy.__call__': ( 'repr_plt.html#plotproxy.__call__',
                                                                                     'lovely_numpy/repr_plt.py'),
                                       'lovely_numpy.repr_plt.PlotProxy.__init__': ( 'repr_plt.html#plotproxy.__init__',
                                                                                     'lovely_numpy/repr_plt.py'),
                                       'lovely_numpy.repr_plt.PlotProxy._repr_png_': ( 'repr_plt.html#plotproxy._repr_png_',
                                                                                       'lovely_numpy/repr_plt.py'),
                                       'lovely_numpy.repr_plt.PlotProxy._repr_svg_': ( 'repr_plt.html#plotproxy._repr_svg_',
                                                                                       'lovely_numpy/repr_plt.py'),
                                       'lovely_numpy.repr_plt.PlotProxy.fig': ('repr_plt.html#plotproxy.fig', 'lovely_numpy/repr_plt.py'),
                                       'lovely_numpy.repr_plt.fig_plot': ('repr_plt.html#fig_plot', 'lovely_numpy/repr_plt.py'),
                                       'lovely_numpy.repr_plt.find_xlims': ('repr_plt.html#find_xlims', 'lovely_numpy/repr_plt.py'),
                                       'lovely_numpy.repr_plt.normal_pdf': ('repr_plt.html#normal_pdf', 'lovely_numpy/repr_plt.py'),
                                       'lovely_numpy.repr_plt.plot': ('repr_plt.html#plot', 'lovely_numpy/repr_plt.py'),
                                       'lovely_numpy.repr_plt.plot_histogram': ('repr_plt.html#plot_histogram', 'lovely_numpy/repr_plt.py'),
                                       'lovely_numpy.repr_plt.plot_minmax': ('repr_plt.html#plot_minmax', 'lovely_numpy/repr_plt.py'),
                                       'lovely_numpy.repr_plt.plot_pdf': ('repr_plt.html#plot_pdf', 'lovely_numpy/repr_plt.py'),
                                       'lovely_numpy.repr_plt.plot_sigmas': ('repr_plt.html#plot_sigmas', 'lovely_numpy/repr_plt.py'),
                                       'lovely_numpy.repr_plt.plot_str': ('repr_plt.html#plot_str', 'lovely_numpy/repr_plt.py'),
                                       'lovely_numpy.repr_plt.sample': ('repr_plt.html#sample', 'lovely_numpy/repr_plt.py')},
            'lovely_numpy.repr_rgb': { 'lovely_numpy.repr_rgb.RGBProxy': ('repr_rgb.html#rgbproxy', 'lovely_numpy/repr_rgb.py'),
                                       'lovely_numpy.repr_rgb.RGBProxy.__call__': ( 'repr_rgb.html#rgbproxy.__call__',
                                                                                    'lovely_numpy/repr_rgb.py'),
                                       'lovely_numpy.repr_rgb.RGBProxy.__init__': ( 'repr_rgb.html#rgbproxy.__init__',
                                                                                    'lovely_numpy/repr_rgb.py'),
                                       'lovely_numpy.repr_rgb.RGBProxy._repr_png_': ( 'repr_rgb.html#rgbproxy._repr_png_',
                                                                                      'lovely_numpy/repr_rgb.py'),
                                       'lovely_numpy.repr_rgb.RGBProxy.fig': ('repr_rgb.html#rgbproxy.fig', 'lovely_numpy/repr_rgb.py'),
                                       'lovely_numpy.repr_rgb.fig_rgb': ('repr_rgb.html#fig_rgb', 'lovely_numpy/repr_rgb.py'),
                                       'lovely_numpy.repr_rgb.rgb': ('repr_rgb.html#rgb', 'lovely_numpy/repr_rgb.py')},
            'lovely_numpy.repr_str': { 'lovely_numpy.repr_str.lovely': ('repr_str.html#lovely', 'lovely_numpy/repr_str.py'),
                                       'lovely_numpy.repr_str.plain_repr': ('repr_str.html#plain_repr', 'lovely_numpy/repr_str.py'),
                                       'lovely_numpy.repr_str.short_dtype': ('repr_str.html#short_dtype', 'lovely_numpy/repr_str.py')},
            'lovely_numpy.utils.colormap': { 'lovely_numpy.utils.colormap.InfCmap': ( 'utils.colormap.html#infcmap',
                                                                                      'lovely_numpy/utils/colormap.py'),
                                             'lovely_numpy.utils.colormap.InfCmap.__call__': ( 'utils.colormap.html#infcmap.__call__',
                                                                                               'lovely_numpy/utils/colormap.py'),
                                             'lovely_numpy.utils.colormap.InfCmap.__init__': ( 'utils.colormap.html#infcmap.__init__',
                                                                                               'lovely_numpy/utils/colormap.py'),
                                             'lovely_numpy.utils.colormap.get_cmap': ( 'utils.colormap.html#get_cmap',
                                                                                       'lovely_numpy/utils/colormap.py')},
            'lovely_numpy.utils.config': { 'lovely_numpy.utils.config.Config': ('utils.config.html#config', 'lovely_numpy/utils/config.py'),
                                           'lovely_numpy.utils.config.Config.__init__': ( 'utils.config.html#config.__init__',
                                                                                          'lovely_numpy/utils/config.py'),
                                           'lovely_numpy.utils.config._Default': ( 'utils.config.html#_default',
                                                                                   'lovely_numpy/utils/config.py'),
                                           'lovely_numpy.utils.config._Default.__repr__': ( 'utils.config.html#_default.__repr__',
                                                                                            'lovely_numpy/utils/config.py'),
                                           'lovely_numpy.utils.config.config': ('utils.config.html#config', 'lovely_numpy/utils/config.py'),
                                           'lovely_numpy.utils.config.get_config': ( 'utils.config.html#get_config',
                                                                                     'lovely_numpy/utils/config.py'),
                                           'lovely_numpy.utils.config.set_config': ( 'utils.config.html#set_config',
                                                                                     'lovely_numpy/utils/config.py')},
            'lovely_numpy.utils.pad': { 'lovely_numpy.utils.pad.pad_frame': ('utils.pad.html#pad_frame', 'lovely_numpy/utils/pad.py'),
                                        'lovely_numpy.utils.pad.pad_frame_gutters': ( 'utils.pad.html#pad_frame_gutters',
                                                                                      'lovely_numpy/utils/pad.py')},
            'lovely_numpy.utils.tile2d': { 'lovely_numpy.utils.tile2d.fit_columns': ( 'utils.tile2d.html#fit_columns',
                                                                                      'lovely_numpy/utils/tile2d.py'),
                                           'lovely_numpy.utils.tile2d.hypertile': ( 'utils.tile2d.html#hypertile',
                                                                                    'lovely_numpy/utils/tile2d.py'),
                                           'lovely_numpy.utils.tile2d.tile2d': ( 'utils.tile2d.html#tile2d',
                                                                                 'lovely_numpy/utils/tile2d.py')},
            'lovely_numpy.utils.utils': { 'lovely_numpy.utils.utils.ansi_color': ( 'utils.utils.html#ansi_color',
                                                                                   'lovely_numpy/utils/utils.py'),
                                          'lovely_numpy.utils.utils.history_warning': ( 'utils.utils.html#history_warning',
                                                                                        'lovely_numpy/utils/utils.py'),
                                          'lovely_numpy.utils.utils.np_to_str_common': ( 'utils.utils.html#np_to_str_common',
                                                                                         'lovely_numpy/utils/utils.py'),
                                          'lovely_numpy.utils.utils.plain_repr': ( 'utils.utils.html#plain_repr',
                                                                                   'lovely_numpy/utils/utils.py'),
                                          'lovely_numpy.utils.utils.plain_str': ( 'utils.utils.html#plain_str',
                                                                                  'lovely_numpy/utils/utils.py'),
                                          'lovely_numpy.utils.utils.pretty_str': ( 'utils.utils.html#pretty_str',
                                                                                   'lovely_numpy/utils/utils.py'),
                                          'lovely_numpy.utils.utils.sci_mode': ('utils.utils.html#sci_mode', 'lovely_numpy/utils/utils.py'),
                                          'lovely_numpy.utils.utils.sparse_join': ( 'utils.utils.html#sparse_join',
                                                                                    'lovely_numpy/utils/utils.py')}}}



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/05_repr_chans.ipynb.

# %% auto 0
__all__ = ['chans']

# %% ../nbs/05_repr_chans.ipynb 3
from typing import Any, Optional as O
from functools import cached_property

import numpy as np
from matplotlib import axes, figure, pyplot as plt
from IPython.core.pylabtools import print_figure

from .repr_rgb import fig_rgb, rgb
from .utils.colormap import InfCmap, get_cmap
from .utils.config import config, get_config

# %% ../nbs/05_repr_chans.ipynb 4
def fig_chans(  x           :np.ndarray,      # Input array
                cmap        :str="twilight",  # Use matplotlib colormap by this name
                cm_below    :str="blue",
                cm_above    :str="red",
                cm_ninf     :str="cyan",
                cm_pinf     :str="fuchsia",
                cm_nan      :str="yellow",
                gutter_px   :int=3,         # Draw write gutters when tiling the images
                frame_px    :int=1,         # Draw black frame around each image
                scale       :int=1,         # Stretch the image. Only itegers please.
                cl          :Any=True,
                view_width  :int=966,
                ax          :O[axes.Axes]=None
        ) -> figure.Figure:
    """
    Process individual channels of a ndarray that can be interpreted as as image
    """
    
    assert x.ndim >= 2, f"Expected a 2+ dim input, got {x.shape}={x.ndim}"
    if x.ndim == 2: x = x[None]
    
    if cl: # Convert to [..., C, H, W].
        x = np.swapaxes(np.swapaxes(x, -2, -1), -3, -2)

    ### XXX Do we want a way to pass a custom cmap instead of mpl one?
    inf_cmap = InfCmap(cmap=get_cmap(cmap),
                  below=cm_below, above=cm_above,
                  nan=cm_nan, ninf=cm_ninf, pinf=cm_pinf)

    return fig_rgb(inf_cmap(x), cl=True, gutter_px=gutter_px, frame_px=frame_px, scale=scale, view_width=view_width, ax=ax)


# %% ../nbs/05_repr_chans.ipynb 5
class ChanProxy():   
    def __init__(self, x: np.ndarray):
        self.x = x
        self.params = dict( cmap        ="twilight", 
                            cm_below    ="blue",
                            cm_above    ="red",
                            cm_ninf     ="cyan",
                            cm_pinf     ="fuchsia",
                            cm_nan      ="yellow",
                            view_width  =966,
                            gutter_px   =3,
                            frame_px    =1,
                            scale       =1,
                            cl          =True,
                            ax          =None)

    def __call__(self,
                 cmap       :O[str] =None, 
                 cm_below   :O[str] =None,
                 cm_above   :O[str] =None,
                 cm_ninf    :O[str] =None,
                 cm_pinf    :O[str] =None,
                 cm_nan     :O[str] =None,
                 view_width :O[int] =None,
                 gutter_px  :O[int] =None,
                 frame_px   :O[int] =None,
                 scale      :O[int] =None,
                 cl         :Any    =None,
                 ax         :O[axes.Axes]=None):
        
        self.params.update( {   k:v for
                                k,v in locals().items()
                                if k != "self" and v is not None } )
        _ = self.fig # Trigger figure generation
        return self

    @cached_property
    def fig(self) -> figure.Figure:
        return fig_chans(self.x, **self.params)

    def _repr_png_(self):
        return print_figure(self.fig, fmt="png", pad_inches=0,
            metadata={"Software": "Matplotlib, https://matplotlib.org/"})


# %% ../nbs/05_repr_chans.ipynb 6
def chans(  x           :np.ndarray,      # Input array
            cmap        :str="twilight",  # Use matplotlib colormap by this name
            cm_below    :str="blue",
            cm_above    :str="red",
            cm_ninf     :str="cyan",
            cm_pinf     :str="fuchsia",
            cm_nan      :str="yellow",
            gutter_px   :int=3,         # Draw write gutters when tiling the images
            frame_px    :int=1,         # Draw black frame around each image
            scale       :int=1,         # Stretch the image. Only itegers please.
            cl          :Any=True,
            view_width  :int=966,
            ax          :O[axes.Axes]=None
        ) -> ChanProxy:
    "Map x values to colors. RGB[A] color is added as channel-last"
    args = locals()
    del args["x"]

    return ChanProxy(x)(**args)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/02_repr_plt.ipynb.

# %% auto 0
__all__ = ['plot']

# %% ../nbs/02_repr_plt.ipynb 3
import math
from functools import cached_property
from typing import Union, Tuple, Any, Optional as O

import numpy as np
from matplotlib import pyplot as plt, axes, figure, rc_context
from IPython.core.pylabtools import print_figure

from .repr_str import lovely, pretty_str
from .utils import get_config, config

# %% ../nbs/02_repr_plt.ipynb 4
def normal_pdf( x   :np.ndarray,
                mean:Union[np.ndarray, float] =0.,
                std :Union[np.ndarray, float] =1.):
    r"""Probability Distribution Function of Normal Distribution:
        $$f(x, \mu, \sigma)
        = \dfrac{1}
            {\sigma \sqrt{2\pi} }
        e^{-\frac{1}{2} \left(\frac{x-\mu}{\sigma}\right)^{2}}
        = \dfrac{e^{- \frac{1}{2}\left(\frac{x-\mu}{\sigma}\right)^{2}}}
            {\sigma \sqrt{2\pi}}$$"""

    np.e
    # dev = x.device
    
    mean = np.array(mean) #if not isinstance(mean, torch.Tensor) else mean
    std = np.array(std) #.to(dev) if not isinstance(std, torch.Tensor) else std

    return (np.array(np.e) ** ( -0.5 * ((x - mean)/std) ** 2 ) /
                (std * np.sqrt((np.pi * 2)))
            )

# %% ../nbs/02_repr_plt.ipynb 6
def sample( x       :np.ndarray,
            max_s   :int,
            plt0    :bool):

    # Samples up to max_s elements and returns
    #   - samples from x
    #   - original x min (None = no good numbes in x)
    #   - original x max (None = no good numbes in x)
    
    # Ignore NaN and Inf.
    x = x[ np.isfinite(x) ]

    x_min = x_max = None

    if x.size: 
        x_min, x_max = x.min(), x.max()
        
        # An option to ignore zeros
        if not plt0: x = x[x != 0.]

        if x.size > max_s and max_s > 0:
            rng = np.random.default_rng( get_config().plt_seed )
            x = rng.choice(x.reshape(-1), max_s) # Sample with replacement for efficiency

    return (x, x_min, x_max)  

# %% ../nbs/02_repr_plt.ipynb 7
def find_xlims( x_min   :Union[float, None],
                x_max   :Union[float, None],
                x_mean  :Union[float, None],
                x_std   :Union[float, None],
                center  :str):

    assert center in ["zero", "mean", "range"]
    
    if x_min is None or x_max is None: return (-1., 1,)
    if x_min == x_max and center == "range": center = "zero"
    if x_mean is None or x_std is None and center == "mean": center = "zero"

    # Center the plot around zero, mean, or the extents of the range of t.
    if center == "range":
        # Center plot arounf the full range of x
        xlim_min, xlim_max = x_min, x_max
    elif center == "mean":
        assert x_mean is not None
        # Center the plot around mean
        # Max distance between mean value and x_min / x_max
        max_div = max(abs(x_mean - x_min), abs(x_max - x_mean))

        xlim_min = x_mean - max_div
        xlim_max = x_mean + max_div
    else: # center == "zero"
        # Center the plot around zero
        abs_max_value = max(abs(x_min), abs(x_max), 1.)
        xlim_min, xlim_max = -abs_max_value, abs_max_value
    

    # Give some extra space around the 
    xlim_min -= abs(xlim_max - xlim_min) * 0.02
    xlim_max += abs(xlim_max - xlim_min) * 0.02

    return (xlim_min, xlim_max)

# %% ../nbs/02_repr_plt.ipynb 8
def plot_histogram( x   :np.ndarray,
                    ax  :axes.Axes):
    if x.size:
        # Around 50 items / bin seems to look good.
        # But don't go below 10 or above 100.
        bins = int(x.size / 50)

        # Adjust the number of bins proportional to the fraction of x axis occupied
        # by the histogram
        xlims = ax.get_xlim()
        
        bins = min(bins, 100)
        bins = np.ceil( bins * ( (x.max()-x.min())/(xlims[1]-xlims[0]) ) ).astype(int)
        bins = max(bins, 10)

        ax.hist(x, bins=bins, color="deepskyblue", align="mid", density=True, zorder=4)

# %% ../nbs/02_repr_plt.ipynb 9
def plot_pdf(   x_mean  :Union[float, None],
                x_std   :Union[float, None],
                ax      :axes.Axes):
    # PDF of normal distribution with the same mean and std.

    if x_std: # Not None and not 0.
        assert x_mean is not None
        xlims = ax.get_xlim()
        xl = np.linspace(*xlims, 100)
        # normal_density = normal_pdf(xl, mean=x_mean, std=x_std)
        ax.plot(xl, normal_pdf(xl, mean=x_mean, std=x_std), zorder=5)


# %% ../nbs/02_repr_plt.ipynb 10
def plot_sigmas(x_min   :Union[float, None],
                x_max   :Union[float, None],
                x_mean  :Union[float, None],
                x_std   :Union[float, None],
                ax      :axes.Axes):
    if x_min is not None and x_max is not None and x_std:
        xlims = ax.get_xlim()
        ylims = ax.get_ylim()
        # Make text bank part of the line under it
        bbox = dict(boxstyle="round", fc="white", edgecolor="none")
        sigmas = max(int(math.floor((abs(x_mean - x_min) / x_std))),
                    int(math.floor((abs(x_max - x_mean) / x_std))))

        for s in range(-sigmas, sigmas+1):
            x_pos = (x_mean + s*x_std)
            if xlims[0] < x_pos < xlims[1]:
                greek = ["-σ", "μ", "+σ"][s+1]  if -1 <= s <= 1 else f"{s:+}σ"
                weight='bold' if not s else None
                ax.axvline(x_pos, 0, 1, c="black")
                ax.text(x_pos, ylims[1]*0.95, greek, ha="center", va="top", bbox=bbox, zorder=5, weight=weight)


# %% ../nbs/02_repr_plt.ipynb 11
def plot_minmax(x_min   :Union[float, None],
                x_max   :Union[float, None],
                ax      :axes.Axes):
    if x_min is not None and x_max is not None:
        bbox = dict(boxstyle="round", fc="white", edgecolor="none", pad=0.)
        y_max = ax.get_ylim()[1]

        # 2 red lines for min and max values
        ax.annotate(
            f"min={pretty_str(x_min)}",
            (x_min, y_max/2),
            xytext=(-1, 0), textcoords='offset points',
            bbox=bbox,
            rotation=90,
            ha="right",
            va="center"
            )

        ax.annotate(
            f"max={pretty_str(x_max)}",
            (x_max, y_max/2),
            xytext=(2, 0), textcoords='offset points',
            bbox=bbox,
            rotation=90,
            ha="left",
            va="center"
            )

        ax.axvline(x_min, 0, 1, c="red", zorder=2)
        ax.axvline(x_max, 0, 1, c="red", zorder=2)


# %% ../nbs/02_repr_plt.ipynb 12
def plot_str(t_str, ax):
    xlim = ax.get_xlim()
    ylim = ax.get_ylim()
    ax.text(xlim[0], ylim[1]*1.05, s=t_str)

# %% ../nbs/02_repr_plt.ipynb 13
def fig_plot(   x     :np.ndarray,  # 
                center  :str    ="zero",        # Center plot on  `zero`, `mean`, or `range`
                max_s   :int    =10000,         # Draw up to this many samples. =0 to draw all
                plt0    :Any    =True,          # Take zero values into account
                ax      :O[axes.Axes]=None,     # Optionally, supply your own matplotlib axes.
                summary :O[str] =None,          # Summary string (to display on top). None=str(lovely(x))
                ddof    :int    =0,             # Apply bias correction to std
        ) -> figure.Figure:
    """Plot statistics"""

    # Useful when you call `plot` from `lovely-tensors`/`-jax` and want to
    # display backend-specific info.
    if summary is None: summary = str(lovely(x, color=False))
    orig_numel = x.size
   
    x, x_min, x_max = sample(x, max_s, plt0)
    x_mean, x_std = (x.mean(), x.std(ddof=ddof)) if x.size else (None,None)


    t_str = ""
    if x.size != orig_numel:
        t_str += str(x.size) 
        if not plt0: t_str += " non-zero"
        t_str += f" samples (μ={pretty_str(x_mean)}, σ={pretty_str(x_std)}) of "
    t_str += summary

    cfg = get_config()
    close = cfg.fig_close and not cfg.fig_show # Don't close if requested to show
    show = cfg.fig_show and ax is None # Don't show the figure if axes was provided

    if not ax:
        fig, ax = plt.subplots(figsize=(12, 2))
        fig.set_constrained_layout(True)
        if close: plt.close(fig)

    xlims = find_xlims(x_min, x_max, x_mean, x_std, center)
    ax.set_xlim(*xlims)
    plot_histogram(x,  ax)
    plot_pdf(x_mean, x_std, ax)
    
    # Add extra space to make sure the labels clear the histogram
    ylim = ax.get_ylim()
    ax.set_ylim( ylim[0], ylim[1]*1.3 )

    plot_sigmas(x_min, x_max, x_mean, x_std, ax)
    plot_minmax(x_min, x_max, ax)
    plot_str(t_str, ax)
    
    ax.set_yticks([])

    if show: plt.show()

    return ax.figure


# %% ../nbs/02_repr_plt.ipynb 14
# This is here for the monkey-patched tensor use case.
# Gives the ability to call both .plt and .plt(ax=ax).  

class PlotProxy(): 
    """Flexible `PIL.Image.Image` wrapper"""

    def __init__(self, x:np.ndarray):
        self.x = x
        self.params = dict( center="zero",
                            max_s=10000,
                            plt0=True,
                            ddof=0,
                            ax=None)

    def __call__(   self,
                    center  :O[str] =None,
                    max_s   :O[int] =None,
                    plt0    :Any    =None,
                    ddof    :O[int] =None,
                    ax      :O[axes.Axes]=None):

        self.params.update( { k:v for
                    k,v in locals().items()
                    if k != "self" and v is not None } )
        
        _ = self.fig # Trigger figure generation
        return self

    @cached_property
    def fig(self) -> figure.Figure:
        return fig_plot( self.x, **self.params)

    def _repr_png_(self):
        return print_figure(self.fig, fmt="png",
            metadata={"Software": "Matplotlib, https://matplotlib.org/"})

    def _repr_svg_(self):
        # Metadata and context for a mode deterministic svg generation
        metadata={
            "Date": None,
            "Creator": "Matplotlib, https://matplotlib.org/",
        }
        with rc_context({"svg.hashsalt": "1"}):
            svg_repr = print_figure(self.fig, fmt="svg", metadata=metadata)
        return svg_repr

# %% ../nbs/02_repr_plt.ipynb 15
def plot(   x       :np.ndarray,        # Your data
            center  :str    ="zero",    # Center plot on  `zero`, `mean`, or `range`
            max_s   :int    =10000,     # Draw up to this many samples. =0 to draw all
            plt0    :Any    =True,      # Take zero values into account
            ax      :O[axes.Axes]=None, # Optionally, supply your own matplotlib axes.
            ddof    :int    =0,         # Apply bias correction to std
        ) -> PlotProxy:

    args = locals()
    del args["x"]

    return PlotProxy(x)(**args)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/03_utils.utils.ipynb.

# %% auto 0
__all__ = ['sci_mode', 'pretty_str', 'plain_repr', 'plain_str', 'sparse_join', 'ansi_color', 'np_to_str_common',
           'history_warning']

# %% ../../nbs/03_utils.utils.ipynb 3
from collections import defaultdict
import warnings
import numpy as np
from typing import Optional, Union
from .config import get_config


# %% ../../nbs/03_utils.utils.ipynb 4
# Do we want this float in decimal or scientific mode?
def sci_mode(f: float):
    config = get_config()
    return ((abs(f) < 10**config.threshold_min) or
            (abs(f) > 10**config.threshold_max))

# %% ../../nbs/03_utils.utils.ipynb 7
# Convert an ndarray or scalar into a string.
# This only looks good for small tensors, which is how it's intended to be used.
def pretty_str(x):
    """A slightly better way to print `float`-y values.
    Works for `np.ndarray`, `torch.Tensor`, `jax.DeviceArray`, and scalars."""

    if isinstance(x, int):
        return '{}'.format(x)
    elif isinstance(x, float):
        if x == 0.:
            return "0."

        sci = sci_mode(x) if get_config().sci_mode is None else get_config().sci_mode
        
        fmt = f"{{:.{get_config().precision}{'e' if sci else 'f'}}}"

        return fmt.format(x)
    elif x.ndim == 0:
            return pretty_str(x.item())
    else:
        slices = [pretty_str(x[i]) for i in range(0, x.shape[0])]
        return '[' + ", ".join(slices) + ']'

# %% ../../nbs/03_utils.utils.ipynb 12
def plain_repr(x):
    "Pick the right function to get a plain repr. `x` is not limited to numpy"
    # assert isinstance(x, np.ndarray), f"expected np.ndarray but got {type(x)}" # Could be a sub-class.
    return x._plain_repr() if hasattr(type(x), "_plain_repr") else repr(x)

def plain_str(x):
    "Pick the right function to get a plain str. `x` is not limited to numpy"
    # assert isinstance(x, np.ndarray), f"expected np.ndarray but got {type(x)}"
    return x._plain_str() if hasattr(type(x), "_plain_str") else str(x)

# %% ../../nbs/03_utils.utils.ipynb 13
def sparse_join(lst, sep=" "):
    # Join non-empty list elements into a space-sepaeated string
    return sep.join( [ l for l in lst if l] )

# %% ../../nbs/03_utils.utils.ipynb 15
def ansi_color(s: str, col: str, use_color=True):
        "Very minimal ANSI color support"
        style = defaultdict(str)
        style["grey"] = "\x1b[38;2;127;127;127m"
        style["red"] = "\x1b[31m"
        end_style = "\x1b[0m"
       
        return style[col]+s+end_style if use_color else s

# %% ../../nbs/03_utils.utils.ipynb 18
def np_to_str_common(x: Union[np.ndarray, np.generic],  # Input
                        color=True,                     # ANSI color highlighting
                        ddof=0):                        # For "std" unbiasing

    if x.size == 0:
        return ansi_color("empty", "grey", color)

    zeros = ansi_color("all_zeros", "grey", color) if np.equal(x, 0.).all() and x.size > 1 else None
    pinf = ansi_color("+Inf!", "red", color) if np.isposinf(x).any() else None
    ninf = ansi_color("-Inf!", "red", color) if np.isneginf(x).any() else None
    nan = ansi_color("NaN!", "red", color) if np.isnan(x).any() else None

    attention = sparse_join([zeros,pinf,ninf,nan])
    numel = f"n={x.size}" if x.size > 5 and max(x.shape) != x.size else None

    summary = None
    if not zeros and isinstance(x, np.ndarray):
        # Calculate stats on good values only.
        gx = x[ np.isfinite(x) ]

        minmax = f"x∈[{pretty_str(gx.min())}, {pretty_str(gx.max())}]" if gx.size > 2 else None
        meanstd = f"μ={pretty_str(gx.mean())} σ={pretty_str(gx.std(ddof=ddof))}" if gx.size >= 2 else None
        summary = sparse_join([numel, minmax, meanstd])


    return sparse_join([ summary, attention])

# %% ../../nbs/03_utils.utils.ipynb 22
def history_warning():
    "Issue a warning (once) ifw e are running in IPYthon with output cache enabled"
    if "get_ipython" in globals() and get_ipython().cache_size > 0:
        warnings.warn("IPYthon has its output cache enabled. See https://xl0.github.io/lovely-tensors/history.html")



#!/usr/bin/env python
# %%
# %matplotlib ipympl

import numpy as np
import matplotlib.pyplot as plt

from lovely_numpy import lo, config, set_config

#%%

numbers = np.load("../nbs/mysteryman.npy").transpose(1,2,0)
in_stats = ( (0.485, 0.456, 0.406),     # mean
             (0.229, 0.224, 0.225) )    # std

mean = np.array(in_stats[0])
std = np.array(in_stats[1])
numbers = (numbers*std + mean).clip(0,1)


#%%

print("make figure, provide axes, plt.show()")

fig = plt.figure(figsize=(8,3))
fig.set_constrained_layout(True)
gs = fig.add_gridspec(2,2)
ax1 = fig.add_subplot(gs[0, :])
ax2 = fig.add_subplot(gs[1, 0])
ax3 = fig.add_subplot(gs[1,1:])

ax2.set_axis_off()
ax3.set_axis_off()

lo(numbers).plt(ax=ax1)
lo(numbers).rgb(ax=ax2)
lo(numbers).chans(ax=ax3)

plt.show();

# %%

print("numers.plt(), plt.show() - nothing should happen")
lo(numbers).plt();
plt.show()

# %%

print("fig_close=False, numers.chans(), plt.show()")
set_config(fig_close=False)
lo(numbers).chans();
plt.show()

#%%
print("fig_show=True, numers.rgb(), no plt.show() - should still show")
set_config(fig_show=True)
lo(numbers).rgb();


# %%



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/03c_utils.tile2d.ipynb.

# %% auto 0
__all__ = ['tile2d', 'hypertile']

# %% ../../nbs/03c_utils.tile2d.ipynb 3
from math import floor, ceil, log2
import numpy as np
from .pad import pad_frame_gutters


# %% ../../nbs/03c_utils.tile2d.ipynb 8
def fit_columns(t: np.ndarray, # Tensor with images, shape=(n,h,w,c)
                view_width=966):
    """Find out how many colums and rows to use to display the images"""
    
    assert t.ndim == 4
    # Let's figure out how many images can we put in a row without the need for
    # re-scaling. Let's try to keep the number as power or 2 if we have to have
    # multiple rows.
    
    n_img = t.shape[0]
    width = t.shape[-2]
    
    if not n_img or not width: return (0, 0)

    n_cols = 2**floor(log2((view_width / width)))

    # At least 1 image per row, even if it does not fit the view without rescaling.
    n_cols = max(1, n_cols)

    # But if we actually don't have enough images to fill a single
    # power-of-two row, just display as many as we got.
    n_cols = min(n_img, n_cols)
    
    n_rows = ceil(n_img / n_cols) # Last row might have free space.
    
    # Avoid producing tilings that are very wide and short.
    while n_rows < n_cols/3:
        n_cols //=2
        n_rows = ceil(n_img / n_cols)

    return (n_rows, n_cols)

# %% ../../nbs/03c_utils.tile2d.ipynb 16
def tile2d(t: np.ndarray,      # Array containing images, shape=(n,h,w,c)
            view_width=966):   # Try to protuce an images at most this wide
    """
    Tile images in a grid.
    """
    assert t.ndim == 4
    # assert t.shape[-1] in (3, 4) # Either RGB or RGBA.

    
    n_images = t.shape[0]
    n_channels = t.shape[-1]
    xy_shape = t.shape[1:3]

    n_rows, n_cols = fit_columns(t, view_width=view_width)

    # We need to form the images inro a rectangular area. For this, we might
    # need to add some dummy images to the last row, whoch might be not be full.
    n_extra_images = n_rows*n_cols - t.shape[0]
    if n_extra_images:
        extra_images = np.ones((n_extra_images, *t.shape[1:]))
        # extra_images = torch.ones((n_extra_images, *t.shape[1:]))
        t = np.concatenate([ t, extra_images ])
    
    # This is where the fun begins! Imagine 't' is tensor[20, 128, 128, 3].
    # and we want 5 rows, 4 columns each.
    
    t = t.reshape(n_rows, n_cols, *t.shape[-3:])
    # Now t is tensor[5, 4, 128, 128, 3]

    t = t.transpose(0, 2, 1, 3, 4)
    # now t is tensor[5, 128, 4, 128, 3]
    # If we just squick dimensions 0,1 and 2,3 togerther, we get the image we want.
    t = t.reshape(n_rows*xy_shape[0], n_cols*xy_shape[1], n_channels)
    
    # Now t is tensor[640, 512, 3], channel-last.
    
    return t

# %% ../../nbs/03c_utils.tile2d.ipynb 21
def hypertile(t: np.ndarray, #torch.Tensor, # input tensor, shape=([...], B, H, W, C)
            frame_px=1,        # Frame width for the innermost group
            gutter_px=3,       # Gutter width for the innermost group
            view_width=966):   # Try to protuce an images at most this wide

    "Recursively tile images on a 2d grid"

    assert t.ndim >= 4, f"Tiling requires at least 3 dimensions: H, W, C. got {t.shape}"
    level = t.ndim - 3

    if t.ndim > 4:
        tlist = list(t)
        sub_view_width = view_width - (gutter_px - frame_px) * level * 2
        tiles = [ hypertile(item,
                            frame_px=frame_px,
                            gutter_px=gutter_px,
                            view_width=sub_view_width)
                  for item in tlist ]
        t = np.stack(tiles)

    return tile2d(pad_frame_gutters(t,
                                     frame_px=frame_px*level,
                                     gutter_px=gutter_px*level),
                    view_width=view_width)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/10_lo.ipynb.

# %% auto 0
__all__ = ['Lo', 'lo']

# %% ../nbs/10_lo.ipynb 3
from typing import Any, Union, Optional as O
from matplotlib import pyplot as plt, axes, figure
from functools import cached_property


import numpy as np
from fastcore.all import store_attr

from .utils import history_warning
from .repr_str import lovely
from .repr_plt import PlotProxy
from .repr_rgb import RGBProxy
from .repr_chans import ChanProxy

# %% ../nbs/10_lo.ipynb 4
class Lo():
    """Lo and behold! What a lovely `numpy.ndarray`!"""
    def __init__(   self,
                    x: Union[np.ndarray, np.generic], # Your data
                    plain       =False, # Show as plain text - values only
                    verbose     =False, # Verbose - show values too
                    depth       =0,     # Expand up to `depth`
                    color :O[bool] =None): # Use ANSI colors
        self.x      =x
        self.plain  =plain
        self.verbose=verbose
        self.depth  =depth
        self.color  =color


        history_warning()

    def __repr__(self):
        return lovely(self.x, plain=self.plain, verbose=self.verbose,
                      depth=self.depth, color=self.color)

    @property
    def v(self):
        "Verbose"
        return Lo(self.x, verbose=True, color=self.color)

    @property
    def p(self):
        "Good old plain representation"
        return Lo(self.x, plain=True, color=self.color)

    @property
    def deeper(self):
        "Going deeper"
        return Lo(self.x, verbose=False, plain=False, depth=1, color=self.color)

    @property
    def rgb(self):
        "Show an image"
        return RGBProxy(self.x)

    @property
    def chans(self):
        "Show color channels"
        return ChanProxy(self.x)

    @property
    def plt(self):
        return PlotProxy(self.x)

    # This is used for .deeper attribute and .deeper(depth=...).
    # The second one results in a __call__.
    def __call__(self, depth=1):
        return Lo(self.x, depth=depth, color=self.color)

# %% ../nbs/10_lo.ipynb 5
def lo(x: Union[np.ndarray, np.generic],    # Your data
        plain   :bool   =False, # Show as plain text - values only
        verbose :bool   =False, # Verbose - show values too
        depth   :int    =0,     # Expand up to `depth`
        color   :O[bool]=None): # Use ANSI colors
    return Lo(x, plain=plain, verbose=verbose, depth=depth, color=color)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_repr_str.ipynb.

# %% auto 0
__all__ = ['lovely']

# %% ../nbs/00_repr_str.ipynb 3
from typing import Union, Optional as O
from collections import defaultdict
from fastcore.foundation import store_attr
import warnings
import numpy as np

from .utils import pretty_str, sparse_join, np_to_str_common
from .utils.config import get_config, set_config, config

# %% ../nbs/00_repr_str.ipynb 5
dtnames =   {   "float16": "f16",
                "float32": "f32",
                "float64": "", # Default dtype in numpy
                "uint8": "u8",
                "uint16": "u16",
                "uint32": "u32",
                "uint64": "u64",
                "int8": "i8",
                "int16": "i16",
                "int32": "i32",
                "int64": "i64",
            }

def short_dtype(x: Union[np.ndarray, np.generic]):
    return dtnames.get(x.dtype.name, str(x.dtype)[6:])

# %% ../nbs/00_repr_str.ipynb 8
def plain_repr(x):
    with config(repr=None):
        return repr(x)

# %% ../nbs/00_repr_str.ipynb 9
def lovely( x       :Union[np.ndarray, np.generic], # The data you want to explore 
            plain   :bool   =False,                 # Plain old way
            verbose :bool   =False,                 # Both summaty and plain
            depth   :int    =0,                     # Show deeper summary, up to `depth`
            lvl     :int    =0,                     # Indentation level 
            color   :O[bool]=None                   # Override `get_config().color`
            ) -> str:                               # The summary

    "Pretty-print the stats of a numpy array or scalar"

    if plain or not isinstance(x, (np.ndarray, np.generic)) or not np.isrealobj(x):
        return plain_repr(x)
    
    if isinstance(x, np.generic):
        tname = None
    else: 
        tname = "array" if type(x) == np.ndarray else type(x).__name__.split(".")[-1]

    shape = str(list(x.shape)) if x.ndim else None
    type_str = sparse_join([tname, shape], sep="")

    color = get_config().color if color is None else color
    common = np_to_str_common(x, color=color)
    dtype = short_dtype(x)
    
    vals = pretty_str(x) if 0 < x.size <= 10 else None
    res = sparse_join([type_str, dtype, common, vals])

    if verbose:
        res += "\n" + plain_repr(x)

    if depth and x.ndim > 1:
        res += "\n" + "\n".join([
            " " * get_config().indent * (lvl+1) +
            str(lovely(x[i,:], depth=depth-1, lvl=lvl+1))
            for i in range(x.shape[0])])

    return res



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/03b_utils.pad.ipynb.

# %% auto 0
__all__ = ['pad_frame', 'pad_frame_gutters']

# %% ../../nbs/03b_utils.pad.ipynb 4
import numpy as np

# %% ../../nbs/03b_utils.pad.ipynb 6
def pad_frame(t: np.ndarray, # torch.Tensor,  # 3D+ image tensor, [...,H,W,C]
              frame_px: int=1,       # Number of pixels to pad each side.
              val :float=0):           # Value to pad with.
    """Pad H and W dimensitons of an image tensor with `val` of thickness `frame_px`"""
    assert t.ndim >= 3

    return np.pad(t, pad_width=[(0,0)] * (t.ndim-3) # Don't pad higher dims
                  + [(frame_px,frame_px)] * 2 #  H and W get padded
                  + [(0,0)], # C does not get padded
                  constant_values=val)


# %% ../../nbs/03b_utils.pad.ipynb 10
def pad_frame_gutters(t: np.ndarray,  # 3D+ Tensor image tensor, [...,H,W,C]
                      gutter_px=3,      # Write gutter in pixels.
                      frame_px=1):      # Black frame, in pixels
    """Add a black frame and white gutters around an image"""
    assert t.ndim >= 3
    xy_shape = t.shape[-2:]
    # gutter_px = ceil(max(xy_shape)*gutter_frac//2)
    
    # XXX This does not work for RGBA images, as the alpha channel is set to 0!
    t = pad_frame(t, frame_px=frame_px, val=0) # Black frame
    return pad_frame(t, frame_px=gutter_px, val=1) # White gutters between images



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/03a_utils.colormap.ipynb.

# %% auto 0
__all__ = ['InfCmap']

# %% ../../nbs/03a_utils.colormap.ipynb 3
from typing import Optional as  O
import numpy as np
import matplotlib as mpl, matplotlib.cm as cm
from matplotlib.colors import Colormap, to_rgba

from ..repr_rgb import rgb

# %% ../../nbs/03a_utils.colormap.ipynb 5
def get_cmap(cmap: str) -> Colormap:
    # Matplotlib changed the colormap interface in version 3.6, and immediately
    # marked the old one as deprecated with a warning. I want to suppot
    # both for the time being, and avoid the warning for people using 3.6+.
    major, minor, *rest = mpl.__version__.split(".")
    assert int(major) == 3 # Drop this compat code when mpl is at 4.0

    if int(minor) <= 5:
        return cm.get_cmap(cmap)
    else:
        return mpl.colormaps[cmap]


# %% ../../nbs/03a_utils.colormap.ipynb 21
class InfCmap():
    """
    Matplotlib colormap extended to have colors for +/-inf

    Parameters extept `cmap` are matplotlib color strings.
    """
    def __init__(self,
                 cmap:  Colormap, # Base matplotlib colormap
                 below: O[str] =None, # Values below 0
                 above: O[str] =None, # Values above 1
                 nan:   O[str] =None, # NaNs
                 ninf:  O[str] =None, # -inf
                 pinf:  O[str] =None, # +inf
                ):
        _ = cmap(0) # one call to make sure the cmap is initialized
        lut = cmap._lut.copy()
        cmax = cmap.N-1
        assert len(lut) == cmap.N+3, "Unexpected colormap LUT size"
        
        
        if below: lut[cmax+1] = np.array(to_rgba(below))
        if above: lut[cmax+2] = np.array(to_rgba(above))
        if nan: lut[cmax+3] = np.array(to_rgba(nan))
        
        # For +/- inf, use above/below as defaults.
        tensor_cmap_ninf = np.array(to_rgba(ninf)) if ninf else lut[cmax+1]
        tensor_cmap_pinf = np.array(to_rgba(pinf)) if pinf else lut[cmax+2]

        # Remove the alpha channel, it causes problems in pad_frame_gutters().
        self.lut = np.concatenate([ lut, tensor_cmap_ninf[None], tensor_cmap_pinf[None] ])[:,:3]
        self.cmax = cmax

    def __call__(self, t: np.ndarray):
        vals = ((t + 1) / 2)
        cmax = self.cmax
        lut_idxs = (vals * cmax).astype(np.int64)
        
        lut_idxs[ vals < 0. ] = cmax+1
        lut_idxs[ vals > 1. ] = cmax+2
        lut_idxs[ np.isnan(t)] = cmax+3
        

        lut_idxs[ np.isneginf(t) ] = cmax+4
        lut_idxs[ np.isposinf(t) ] = cmax+5
        
        return self.lut.take(lut_idxs, axis=0, mode="clip") # RGB added as color-last.         
